from json_config_expander.cartesian_product_config_expander import expand_configs as cartesian_product_expand_configs
from json_config_expander.find_and_replace_config_expander import expand_configs as find_and_replace_expand_config


class ImplementationTypes:
	CARTESIAN_PRODUCT_BASED = cartesian_product_expand_configs
	FIND_AND_REPLACE_BASED = find_and_replace_expand_config


def expand_configs(base_config, function=None, expand_char='*',
		expand_configs_implementation=ImplementationTypes.CARTESIAN_PRODUCT_BASED):
	return expand_configs_implementation(base_config, function, expand_char)



from copy import deepcopy, copy
from itertools import product, chain


def seq_iter(obj):
	# iterate dict and list in same way
	return obj if isinstance(obj, dict) else range(len(obj))


class expand_configs(object):
	def __new__(cls, base_config, function=None, expand_char='*'):
		self = object.__new__(cls)
		self.expand_char = expand_char
		return self(base_config, function)

	def __call__(self, base_config, function=None):
		expanded_configs = self._expand_configs(base_config)
		if function is not None:
			return [function(deepcopy(config)) for config in expanded_configs]
		else:
			return [deepcopy(config) for config in expanded_configs]

	def _expand_configs(self, base_config):
		"""
		For each child we calculate its versions recursively and create new config with all possible combinations
		"""
		if not isinstance(base_config, dict) and not isinstance(base_config, list):
			return [base_config]
		expanded_config = self._expand_config_children(base_config)
		return self._create_cartesian_product_from_children(expanded_config)

	def _expand_config_children(self, config):
		expanded_config = copy(config)
		for key in seq_iter(expanded_config):
			item = expanded_config[key]
			if isinstance(key, str) and self.expand_char in key:
				if not isinstance(item, list): raise TypeError("You can apply expand only on list")
				# expand each child of item and flatten results to single list
				expanded_config[key] = chain.from_iterable(self._expand_configs(x) for x in item)
			else:
				expanded_config[key] = self._expand_configs(item)
		return expanded_config

	def _create_cartesian_product_from_children(self, expanded_config):
		if isinstance(expanded_config, list):
			# convert items to list, since product returns tuples
			return [list(x) for x in product(*expanded_config)]

		# create dicts from config keys and config product values
		keys = [k.replace(self.expand_char, '') for k in expanded_config.keys()]
		return list(dict(zip(keys, x)) for x in product(*expanded_config.values()))



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
	return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
	name="json-config-expander",
	version="0.6",
	author="Diagnostic Robotics",
	author_email="admin@diagnosticrobotics.com",
	description="Expand multi optional configuration to multiple configurations.",
	license="MIT",
	keywords="multiple json config",
	url="https://github.com/DiagnosticRobotics/json-config-expander",
	packages=find_packages(include=['json_config_expander', 'json_config_expander.*']),
	long_description=read('README.md'),
	long_description_content_type='text/markdown',
	classifiers=["Development Status :: 3 - Alpha"]
)



import copy


class expand_configs(object):
	def __new__(cls, base_config, function=None, expand_char='*'):
		self = object.__new__(cls)
		self.expand_char = expand_char
		self.type_handle_mapping = {dict: self._handle_dict, list: self._handle_list}
		return self(base_config, function)

	def __call__(self, base_config, function=None):
		if function is not None:
			expanded_configs = self(base_config)
			return [function(config) for config in expanded_configs]

		path = self._find_first_expand_char_path(base_config)
		if path is None:
			return [base_config]

		path_value = self._get_path_value(base_config, path)

		expanded_configs = []
		for value in path_value:
			new_value = copy.deepcopy(value)
			new_config = copy.deepcopy(base_config)
			self._update_value_in_path(new_config, new_value, path)
			expanded_configs.extend(self(new_config))

		return expanded_configs

	def _find_first_expand_char_path(self, base_config):
		base_config_type = type(base_config)
		if base_config_type in self.type_handle_mapping:
			return self.type_handle_mapping[base_config_type](base_config)
		return None

	def _handle_list(self, config_list):
		for index in range(len(config_list)):
			path = self._find_first_expand_char_path(config_list[index])
			if path is not None:
				return [index] + path
		return None

	def _handle_dict(self, config_dict):
		for key in config_dict:
			if self.expand_char in key:
				return [key]

			path = self._find_first_expand_char_path(config_dict[key])
			if path is not None:
				return [key] + path
		return None

	def _update_value_in_path(self, new_config, new_value, path):
		path_value = new_config
		for key_index in range(len(path) - 1):
			key = path[key_index]
			path_value = path_value[key]
		last_key = path[-1]  # The last key must be a string
		key_without_expand_char = last_key.replace(self.expand_char, '')
		path_value[key_without_expand_char] = new_value
		path_value.pop(last_key)

	def _get_path_value(self, base_config, path):
		path_value = base_config
		for key in path:
			path_value = path_value[key]
		if type(path_value) is not list:
			raise TypeError('the value of {} should be a list! ({} is not a list)'.format(path, path_value))
		return path_value
